@@ -86,7 +86,7 @@ struct _GtkBindingSignal
 
 struct _GtkBindingArg
 {
-  GtkType	arg_type;
+  GType		 arg_type;
   union {
     glong	 long_data;
     gdouble	 double_data;
@@ -103,6 +103,8 @@ GtkBindingSet*	gtk_binding_set_find	(const gchar	*set_name);
 gboolean gtk_bindings_activate		(GtkObject	*object,
 					 guint		 keyval,
 					 GdkModifierType modifiers);
+gboolean gtk_bindings_activate_event    (GtkObject      *object,
+					 GdkEventKey    *event);
 gboolean gtk_binding_set_activate	(GtkBindingSet	*binding_set,
 					 guint		 keyval,
 					 GdkModifierType modifiers,
@@ -136,8 +138,21 @@ void	 gtk_binding_entry_add_signall	(GtkBindingSet	*binding_set,
 guint	 gtk_binding_parse_binding	(GScanner	*scanner);
 
 
-void     _gtk_binding_reset_parsed      (void);
+void     _gtk_binding_reset_parsed    (void);
 
+/* Creates a signal with a fixed callback instead of a class offset;
+ * useful for key binding signals
+ */
+guint _gtk_binding_signal_new (const gchar       *signal_name,
+			       GType		  itype,
+			       GSignalFlags	  signal_flags,
+			       GCallback          handler,
+			       GSignalAccumulator accumulator,
+			       gpointer		  accu_data,
+			       GSignalCMarshaller c_marshaller,
+			       GType		  return_type,
+			       guint		  n_params,
+			       ...);
 #ifdef __cplusplus
 }
 #endif /* __cplusplus */
