@@ -385,6 +385,8 @@ gtk_image_set_property (GObject      *object,
     case PROP_ICON_NAME:
       gtk_image_set_from_icon_name (image, g_value_get_string (value),
 				    image->icon_size);
+      break;
+
     default:
       G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
       break;
@@ -2106,3 +2108,35 @@ gtk_image_get_pixel_size (GtkImage *image)
 
   return priv->pixel_size;
 }
+
+#ifdef G_OS_WIN32
+
+#undef gtk_image_new_from_file
+
+GtkWidget*
+gtk_image_new_from_file   (const gchar *filename)
+{
+  gchar *utf8_filename = g_locale_to_utf8 (filename, -1, NULL, NULL, NULL);
+  GtkWidget *retval;
+
+  retval = gtk_image_new_from_file_utf8 (utf8_filename);
+
+  g_free (utf8_filename);
+
+  return retval;
+}
+
+#undef gtk_image_set_from_file
+
+void
+gtk_image_set_from_file   (GtkImage    *image,
+                           const gchar *filename)
+{
+  gchar *utf8_filename = g_locale_to_utf8 (filename, -1, NULL, NULL, NULL);
+
+  gtk_image_set_from_file_utf8 (image, utf8_filename);
+
+  g_free (utf8_filename);
+}
+
+#endif
