@@ -25,22 +25,15 @@
  */
 
 #include <config.h>
+#include "gtkalias.h"
 #include "gtkimagemenuitem.h"
 #include "gtkaccellabel.h"
 #include "gtkintl.h"
 #include "gtkstock.h"
 #include "gtkiconfactory.h"
 #include "gtkimage.h"
-
-typedef struct _GtkImageMenuItemPrivate GtkImageMenuItemPrivate;
-
-struct _GtkImageMenuItemPrivate 
-{
-  GtkSettings *settings;
-  guint        show_image_connection;
-};
-
-#define GTK_IMAGE_MENU_ITEM_GET_PRIVATE(o)  (G_TYPE_INSTANCE_GET_PRIVATE ((o), GTK_TYPE_IMAGE_MENU_ITEM, GtkImageMenuItemPrivate))
+#include "gtkcontainer.h"
+#include "gtkwindow.h"
 
 static void gtk_image_menu_item_class_init           (GtkImageMenuItemClass *klass);
 static void gtk_image_menu_item_init                 (GtkImageMenuItem      *image_menu_item);
@@ -145,8 +138,7 @@ gtk_image_menu_item_class_init (GtkImageMenuItemClass *klass)
 						       P_("Whether images should be shown in menus"),
 						       TRUE,
 						       G_PARAM_READWRITE));
-
-  g_type_class_add_private (gobject_class, sizeof (GtkImageMenuItemPrivate));  
+  
 }
 
 static void
@@ -220,16 +212,19 @@ gtk_image_menu_item_toggle_size_request (GtkMenuItem *menu_item,
 
   if (image_menu_item->image && show_image (image_menu_item))
     {
-      guint image_width = image_menu_item->image->requisition.width;
+      GtkRequisition image_requisition;
 
-      if (image_width > 0)
+      gtk_widget_get_child_requisition (image_menu_item->image,
+                                        &image_requisition);
+
+      if (image_requisition.width > 0)
 	{
 	  guint toggle_spacing;
 	  gtk_widget_style_get (GTK_WIDGET (menu_item),
 				"toggle_spacing", &toggle_spacing,
 				NULL);
-	  
-	  *requisition = image_width + toggle_spacing;
+
+	  *requisition = image_requisition.width + toggle_spacing;
 	}
     }
 }
@@ -281,7 +276,8 @@ gtk_image_menu_item_size_allocate (GtkWidget     *widget,
 
   if (image_menu_item->image && show_image (image_menu_item))
     {
-      gint width, height, x, y, offset;
+      gint x, y, offset;
+      GtkRequisition child_requisition;
       GtkAllocation child_allocation;
       guint horizontal_padding, toggle_spacing;
 
@@ -293,9 +289,10 @@ gtk_image_menu_item_size_allocate (GtkWidget     *widget,
       /* Man this is lame hardcoding action, but I can't
        * come up with a solution that's really better.
        */
-      
-      width = image_menu_item->image->requisition.width;
-      height = image_menu_item->image->requisition.height;
+
+      gtk_widget_get_child_requisition (image_menu_item->image,
+                                        &child_requisition);
+
       offset = GTK_CONTAINER (image_menu_item)->border_width +
 	widget->style->xthickness;
 
@@ -303,20 +300,20 @@ gtk_image_menu_item_size_allocate (GtkWidget     *widget,
 	{
 	  x = offset + horizontal_padding +
 	      (GTK_MENU_ITEM (image_menu_item)->toggle_size -
-	       toggle_spacing - width) / 2;
+	       toggle_spacing - child_requisition.width) / 2;
 	}
       else
 	{
 	  x = widget->allocation.width - offset - horizontal_padding -
 	      GTK_MENU_ITEM (image_menu_item)->toggle_size + toggle_spacing +
 	      (GTK_MENU_ITEM (image_menu_item)->toggle_size -
-	       toggle_spacing - width) / 2;
+	       toggle_spacing - child_requisition.width) / 2;
 	}
 
-      y = (widget->allocation.height - height) / 2;
+      y = (widget->allocation.height - child_requisition.height) / 2;
 
-      child_allocation.width = width;
-      child_allocation.height = height;
+      child_allocation.width = child_requisition.width;
+      child_allocation.height = child_requisition.height;
       child_allocation.x = widget->allocation.x + MAX (x, 0);
       child_allocation.y = widget->allocation.y + MAX (y, 0);
 
@@ -470,6 +467,8 @@ gtk_image_menu_item_new_from_stock (const gchar      *stock_id,
  * @image: a widget to set as the image for the menu item.
  * 
  * Sets the image of @image_menu_item to the given widget.
+ * Note that it depends on the show-menu-images setting whether
+ * the image will be displayed or not.
  **/ 
 void
 gtk_image_menu_item_set_image (GtkImageMenuItem *image_menu_item,
@@ -545,44 +544,65 @@ gtk_image_menu_item_remove (GtkContainer *container,
 static void 
 show_image_change_notify (GtkImageMenuItem *image_menu_item)
 {
-  if (image_menu_item->image) 
-    g_object_set (image_menu_item->image, "visible", show_image (image_menu_item), NULL);
+  if (image_menu_item->image)
+    {
+      if (show_image (image_menu_item))
+	gtk_widget_show (image_menu_item->image);
+      else
+	gtk_widget_hide (image_menu_item->image);
+    }
+}
+
+static void
+traverse_container (GtkWidget *widget,
+		    gpointer   data)
+{
+  if (GTK_IS_IMAGE_MENU_ITEM (widget))
+    show_image_change_notify (GTK_IMAGE_MENU_ITEM (widget));
+  else if (GTK_IS_CONTAINER (widget))
+    gtk_container_forall (GTK_CONTAINER (widget), traverse_container, NULL);
+}
+
+static void
+gtk_image_menu_item_setting_changed (GtkSettings *settings)
+{
+  GList *list, *l;
+
+  list = gtk_window_list_toplevels ();
+
+  for (l = list; l; l = l->next)
+    gtk_container_forall (GTK_CONTAINER (l->data), 
+			  traverse_container, NULL);
+
+  g_list_free (list);  
 }
 
 static void
 gtk_image_menu_item_screen_changed (GtkWidget *widget,
 				    GdkScreen *previous_screen)
 {
-  GtkImageMenuItemPrivate *priv = GTK_IMAGE_MENU_ITEM_GET_PRIVATE (widget);
   GtkSettings *settings;
+  guint show_image_connection;
 
-  if (gtk_widget_has_screen (widget))
-    settings = gtk_widget_get_settings (widget);
-  else
-    settings = NULL;
-
-  if (settings == priv->settings)
+  if (!gtk_widget_has_screen (widget))
     return;
 
-  if (priv->settings)
-    {
-      g_signal_handler_disconnect (priv->settings, priv->show_image_connection);
-      g_object_unref (priv->settings);
-    }
+  settings = gtk_widget_get_settings (widget);
+  
+  show_image_connection = 
+    GPOINTER_TO_UINT (g_object_get_data (G_OBJECT (settings), 
+					 "gtk-image-menu-item-connection"));
+  
+  if (show_image_connection)
+    return;
 
-  if (settings)
-    {
-      priv->show_image_connection =
-	g_signal_connect_swapped (settings,
-				  "notify::gtk-menu-images",
-				  G_CALLBACK (show_image_change_notify),
-				  widget);
-
-      g_object_ref (settings);
-      priv->settings = settings;
-    }
-  else
-    priv->settings = NULL;
+  show_image_connection =
+    g_signal_connect (settings, "notify::gtk-menu-images",
+		      G_CALLBACK (gtk_image_menu_item_setting_changed), 0);
+  g_object_set_data (G_OBJECT (settings), 
+		     "gtk-image-menu-item-connection",
+		     GUINT_TO_POINTER (show_image_connection));
 
   show_image_change_notify (GTK_IMAGE_MENU_ITEM (widget));
 }
+
