@@ -32,8 +32,10 @@
 #include <stdlib.h>
 #include <string.h>
 
+#include <glib/gstdio.h>
 #include <gmodule.h>
 #include <pango/pango-utils.h>
+#include "gtkalias.h"
 #include "gtkimmodule.h"
 #include "gtkimcontextsimple.h"
 #include "gtkrc.h"
@@ -44,6 +46,10 @@
  * else.
  */
 
+#ifdef __GTK_PRIVATE_H__
+#error gtkprivate.h should not be included in this file
+#endif
+
 #define SIMPLE_ID "gtk-im-context-simple"
 
 typedef struct _GtkIMModule      GtkIMModule;
@@ -76,7 +82,7 @@ struct _GtkIMModuleClass
   GTypeModuleClass parent_class;
 };
 
-GType gtk_im_module_get_type (void);
+static GType gtk_im_module_get_type (void);
 
 static gint n_loaded_contexts = 0;
 static GHashTable *contexts_hash = NULL;
@@ -162,7 +168,7 @@ gtk_im_module_class_init (GtkIMModuleClass *class)
   gobject_class->finalize = gtk_im_module_finalize;
 }
 
-GType
+static GType
 gtk_im_module_get_type (void)
 {
   static GType im_module_type = 0;
@@ -265,7 +271,7 @@ correct_libdir_prefix (gchar **path)
 
 
 static void
-gtk_im_module_init ()
+gtk_im_module_init (void)
 {
   GString *line_buf = g_string_new (NULL);
   GString *tmp_buf = g_string_new (NULL);
@@ -278,7 +284,7 @@ gtk_im_module_init ()
 
   contexts_hash = g_hash_table_new (g_str_hash, g_str_equal);
 
-  file = fopen (filename, "r");
+  file = g_fopen (filename, "r");
   if (!file)
     {
       /* In case someone wants only the default input method,
@@ -352,7 +358,7 @@ gtk_im_module_init ()
 	    goto context_error;
 	  info->domain_dirname = g_strdup (tmp_buf->str);
 #ifdef DO_CORRECT_LIBDIR_PREFIX
-	  correct_libdir_prefix (&info->domain_dirname);
+	  correct_libdir_prefix ((char **) &info->domain_dirname);
 #endif
 
 	  if (!pango_scan_string (&p, tmp_buf))
