@@ -24,7 +24,9 @@
  * GTK+ at ftp://ftp.gtk.org/pub/gtk/. 
  */
 
+#include <config.h>
 #include <gdk/gdk.h>
+#include "gtkalias.h"
 #include "gtkinvisible.h"
 #include "gtkintl.h"
 
