@@ -26,6 +26,7 @@
  */
 
 #include <config.h>
+#include "gtkalias.h"
 #include "gtkframe.h"
 #include "gtklabel.h"
 #include "gtkmarshalers.h"
@@ -87,6 +88,9 @@ static void     gtk_statusbar_set_property      (GObject           *object,
 						 guint              prop_id,
 						 const GValue      *value,
 						 GParamSpec        *pspec);
+static void     label_selectable_changed        (GtkWidget         *label,
+                                                 GParamSpec        *pspec,
+						 gpointer           data);
 
 
 static GtkContainerClass *parent_class;
@@ -225,11 +229,11 @@ gtk_statusbar_init (GtkStatusbar *statusbar)
   gtk_widget_show (statusbar->frame);
 
   statusbar->label = gtk_label_new ("");
+  gtk_label_set_single_line_mode (GTK_LABEL (statusbar->label), TRUE);
   gtk_misc_set_alignment (GTK_MISC (statusbar->label), 0.0, 0.5);
-  /* don't expand the size request for the label; if we
-   * do that then toplevels weirdly resize
-   */
-  gtk_widget_set_size_request (statusbar->label, 1, -1);
+  g_signal_connect (statusbar->label, "notify::selectable",
+		    G_CALLBACK (label_selectable_changed), statusbar);
+  gtk_label_set_ellipsize (GTK_LABEL (statusbar->label), PANGO_ELLIPSIZE_END);
   gtk_container_add (GTK_CONTAINER (statusbar->frame), statusbar->label);
   gtk_widget_show (statusbar->label);
 
@@ -409,16 +413,21 @@ gtk_statusbar_set_has_resize_grip (GtkStatusbar *statusbar,
   if (setting != statusbar->has_resize_grip)
     {
       statusbar->has_resize_grip = setting;
+      gtk_widget_queue_resize (statusbar->label);
       gtk_widget_queue_draw (GTK_WIDGET (statusbar));
 
       if (GTK_WIDGET_REALIZED (statusbar))
         {
           if (statusbar->has_resize_grip && statusbar->grip_window == NULL)
-            gtk_statusbar_create_window (statusbar);
+	    {
+	      gtk_statusbar_create_window (statusbar);
+	      if (GTK_WIDGET_MAPPED (statusbar))
+		gdk_window_show (statusbar->grip_window);
+	    }
           else if (!statusbar->has_resize_grip && statusbar->grip_window != NULL)
             gtk_statusbar_destroy_window (statusbar);
         }
-
+      
       g_object_notify (G_OBJECT (statusbar), "has_resize_grip");
     }
 }
@@ -524,10 +533,10 @@ get_grip_rect (GtkStatusbar *statusbar,
   w = 18;
   h = 18;
 
-  if (w > (widget->allocation.width))
+  if (w > widget->allocation.width)
     w = widget->allocation.width;
 
-  if (h > (widget->allocation.height - widget->style->ythickness))
+  if (h > widget->allocation.height - widget->style->ythickness)
     h = widget->allocation.height - widget->style->ythickness;
   
   rect->width = w;
@@ -589,6 +598,7 @@ gtk_statusbar_create_window (GtkStatusbar *statusbar)
 
   statusbar->grip_window = gdk_window_new (widget->window,
                                            &attributes, attributes_mask);
+
   gdk_window_set_user_data (statusbar->grip_window, widget);
 
   set_grip_cursor (statusbar);
@@ -674,7 +684,8 @@ gtk_statusbar_button_press (GtkWidget      *widget,
   statusbar = GTK_STATUSBAR (widget);
   
   if (!statusbar->has_resize_grip ||
-    event->type != GDK_BUTTON_PRESS)
+      event->type != GDK_BUTTON_PRESS ||
+      event->window != statusbar->grip_window)
     return FALSE;
   
   ancestor = gtk_widget_get_toplevel (widget);
@@ -751,53 +762,112 @@ gtk_statusbar_size_request   (GtkWidget      *widget,
   gtk_frame_set_shadow_type (GTK_FRAME (statusbar->frame), shadow_type);
   
   GTK_WIDGET_CLASS (parent_class)->size_request (widget, requisition);
+}
 
-  if (statusbar->has_resize_grip)
+/* look for extra children between the frame containing
+ * the label and where we want to draw the resize grip 
+ */
+static gboolean
+has_extra_children (GtkStatusbar *statusbar)
+{
+  GList *l;
+  GtkBoxChild *child, *frame;
+
+  frame = NULL;
+  for (l = GTK_BOX (statusbar)->children; l; l = l->next)
     {
-      GdkRectangle rect;
+      frame = l->data;
 
-      /* x, y in the grip rect depend on size allocation, but
-       * w, h do not so this is OK
-       */
-      get_grip_rect (statusbar, &rect);
-      
-      requisition->width += rect.width;
-      requisition->height = MAX (requisition->height, rect.height);
+      if (frame->widget == statusbar->frame)
+	break;
     }
+  
+  for (l = l->next; l; l = l->next)
+    {
+      child = l->data;
+
+      if (!GTK_WIDGET_VISIBLE (child->widget))
+	continue;
+
+      if (frame->pack == GTK_PACK_START || child->pack == GTK_PACK_END)
+	return TRUE;
+    }
+
+  return FALSE;
 }
 
 static void
 gtk_statusbar_size_allocate  (GtkWidget     *widget,
                               GtkAllocation *allocation)
 {
-  GtkStatusbar *statusbar;
-  
-  statusbar = GTK_STATUSBAR (widget);
+  GtkStatusbar *statusbar = GTK_STATUSBAR (widget);
+  gboolean extra_children = FALSE;
+  GdkRectangle rect;
 
   if (statusbar->has_resize_grip)
     {
-      GdkRectangle rect;
-      GtkRequisition saved_req;
+      widget->allocation = *allocation;
+      get_grip_rect (statusbar, &rect);    
       
-      widget->allocation = *allocation; /* get_grip_rect needs this info */
-      get_grip_rect (statusbar, &rect);
-  
-      if (statusbar->grip_window)
-        gdk_window_move_resize (statusbar->grip_window,
-                                rect.x, rect.y,
-                                rect.width, rect.height);
+      extra_children = has_extra_children (statusbar);
       
-      /* enter the bad hack zone */      
-      saved_req = widget->requisition;
-      widget->requisition.width -= rect.width; /* HBox::size_allocate needs this */
-      if (widget->requisition.width < 0)
-        widget->requisition.width = 0;
-      GTK_WIDGET_CLASS (parent_class)->size_allocate (widget, allocation);
-      widget->requisition = saved_req;
+      /* If there are extra children, we don't want them to occupy
+       * the space where we draw the resize grip, so we temporarily
+       * shrink the allocation.
+       * If there are no extra children, we want the frame to get
+       * the full allocation, and we fix up the allocation of the
+       * label afterwards to make room for the grip.
+       */
+      if (extra_children)
+	{
+	  allocation->width -= rect.width;
+	  if (gtk_widget_get_direction (widget) == GTK_TEXT_DIR_RTL) 
+	    allocation->x += rect.width;
+	}
     }
-  else
+      
+  /* chain up normally */
+  GTK_WIDGET_CLASS (parent_class)->size_allocate (widget, allocation);
+
+  if (statusbar->has_resize_grip)
     {
-      /* chain up normally */
-      GTK_WIDGET_CLASS (parent_class)->size_allocate (widget, allocation);
+      if (statusbar->grip_window)
+	{
+	  gdk_window_raise (statusbar->grip_window);
+	  gdk_window_move_resize (statusbar->grip_window,
+				  rect.x, rect.y,
+				  rect.width, rect.height);
+	}
+
+      if (extra_children) 
+	{
+	  allocation->width += rect.width;
+	  if (gtk_widget_get_direction (widget) == GTK_TEXT_DIR_RTL) 
+	    allocation->x -= rect.width;
+	  
+	  widget->allocation = *allocation;
+	}
+      else
+	{
+	  /* shrink the label to make room for the grip */
+	  *allocation = statusbar->label->allocation;
+	  allocation->width -= rect.width;
+	  if (gtk_widget_get_direction (widget) == GTK_TEXT_DIR_RTL) 
+	    allocation->x += rect.width;
+	  
+	  gtk_widget_size_allocate (statusbar->label, allocation);
+	}
     }
 }
+
+static void
+label_selectable_changed (GtkWidget  *label,
+			  GParamSpec *pspec,
+			  gpointer    data)
+{
+  GtkStatusbar *statusbar = GTK_STATUSBAR (data);
+
+  if (statusbar && 
+      statusbar->has_resize_grip && statusbar->grip_window)
+    gdk_window_raise (statusbar->grip_window);
+}
