@@ -37,11 +37,11 @@ extern "C" {
 #endif /* __cplusplus */
 
 #define GTK_TYPE_TEXT_VIEW             (gtk_text_view_get_type ())
-#define GTK_TEXT_VIEW(obj)             (GTK_CHECK_CAST ((obj), GTK_TYPE_TEXT_VIEW, GtkTextView))
-#define GTK_TEXT_VIEW_CLASS(klass)     (GTK_CHECK_CLASS_CAST ((klass), GTK_TYPE_TEXT_VIEW, GtkTextViewClass))
-#define GTK_IS_TEXT_VIEW(obj)          (GTK_CHECK_TYPE ((obj), GTK_TYPE_TEXT_VIEW))
-#define GTK_IS_TEXT_VIEW_CLASS(klass)  (GTK_CHECK_CLASS_TYPE ((klass), GTK_TYPE_TEXT_VIEW))
-#define GTK_TEXT_VIEW_GET_CLASS(obj)   (GTK_CHECK_GET_CLASS ((obj), GTK_TYPE_TEXT_VIEW, GtkTextViewClass))
+#define GTK_TEXT_VIEW(obj)             (G_TYPE_CHECK_INSTANCE_CAST ((obj), GTK_TYPE_TEXT_VIEW, GtkTextView))
+#define GTK_TEXT_VIEW_CLASS(klass)     (G_TYPE_CHECK_CLASS_CAST ((klass), GTK_TYPE_TEXT_VIEW, GtkTextViewClass))
+#define GTK_IS_TEXT_VIEW(obj)          (G_TYPE_CHECK_INSTANCE_TYPE ((obj), GTK_TYPE_TEXT_VIEW))
+#define GTK_IS_TEXT_VIEW_CLASS(klass)  (G_TYPE_CHECK_CLASS_TYPE ((klass), GTK_TYPE_TEXT_VIEW))
+#define GTK_TEXT_VIEW_GET_CLASS(obj)   (G_TYPE_INSTANCE_GET_CLASS ((obj), GTK_TYPE_TEXT_VIEW, GtkTextViewClass))
 
 typedef enum
 {
@@ -89,12 +89,14 @@ struct _GtkTextView
   
   guint overwrite_mode : 1;
   guint cursor_visible : 1;
-  guint  need_im_reset : 1;	/* If we have reset the IM since the last character entered */
-  /* just selected a word or line via double/triple click */
-  guint just_selected_element : 1;
 
-  /* disable scrolling to cursor on focus */
-  guint disable_scroll_on_focus : 1;
+  /* if we have reset the IM since the last character entered */  
+  guint  need_im_reset : 1;	
+
+  guint accepts_tab : 1;
+  
+  /* this flag is no longer used */
+  guint reserved : 1;
   
   /* debug flag - means that we've validated onscreen since the
    * last "invalidate" signal from the layout
@@ -147,7 +149,6 @@ struct _GtkTextView
   GSList *children;
 
   GtkTextPendingScroll *pending_scroll;
-  GdkGC *cursor_gc;
 
   gint pending_place_cursor_button;
 };
@@ -188,6 +189,7 @@ struct _GtkTextViewClass
   void (* delete_from_cursor)    (GtkTextView  *text_view,
                                   GtkDeleteType type,
                                   gint          count);
+  void (* backspace)             (GtkTextView *text_view);
 
   /* cut copy paste */
   void (* cut_clipboard)   (GtkTextView *text_view);
@@ -209,10 +211,9 @@ struct _GtkTextViewClass
   void (*_gtk_reserved5) (void);
   void (*_gtk_reserved6) (void);
   void (*_gtk_reserved7) (void);
-  void (*_gtk_reserved8) (void);
 };
 
-GtkType        gtk_text_view_get_type              (void) G_GNUC_CONST;
+GType          gtk_text_view_get_type              (void) G_GNUC_CONST;
 GtkWidget *    gtk_text_view_new                   (void);
 GtkWidget *    gtk_text_view_new_with_buffer       (GtkTextBuffer *buffer);
 void           gtk_text_view_set_buffer            (GtkTextView   *text_view,
@@ -249,6 +250,11 @@ void           gtk_text_view_get_iter_at_location  (GtkTextView   *text_view,
                                                     GtkTextIter   *iter,
                                                     gint           x,
                                                     gint           y);
+void           gtk_text_view_get_iter_at_position  (GtkTextView   *text_view,
+                                                    GtkTextIter   *iter,
+						    gint          *trailing,
+                                                    gint           x,
+                                                    gint           y);
 void           gtk_text_view_get_line_yrange       (GtkTextView       *text_view,
                                                     const GtkTextIter *iter,
                                                     gint              *y,
@@ -323,6 +329,12 @@ GtkWrapMode      gtk_text_view_get_wrap_mode          (GtkTextView      *text_vi
 void             gtk_text_view_set_editable           (GtkTextView      *text_view,
                                                        gboolean          setting);
 gboolean         gtk_text_view_get_editable           (GtkTextView      *text_view);
+void             gtk_text_view_set_overwrite          (GtkTextView      *text_view,
+						       gboolean          overwrite);
+gboolean         gtk_text_view_get_overwrite          (GtkTextView      *text_view);
+void		 gtk_text_view_set_accepts_tab        (GtkTextView	*text_view,
+						       gboolean		 accepts_tab);
+gboolean	 gtk_text_view_get_accepts_tab        (GtkTextView	*text_view);
 void             gtk_text_view_set_pixels_above_lines (GtkTextView      *text_view,
                                                        gint              pixels_above_lines);
 gint             gtk_text_view_get_pixels_above_lines (GtkTextView      *text_view);
