@@ -21,6 +21,8 @@
  * Authors: Alexander Larsson <alexl@redhat.com>
  */
 
+#include <config.h>
+#include "gtkalias.h"
 #include "gtkprivate.h"
 #include "gtkwindow.h"
 #include "gtkmain.h"
