@@ -95,6 +95,13 @@ no_backup_files_filter (const GtkFileFilterInfo *filter_info,
     return 1;
 }
 
+static void
+filter_changed (GtkFileChooserDialog *dialog,
+		gpointer              data)
+{
+  g_print ("file filter changed\n");
+}
+
 static char *
 format_time (time_t t)
 {
@@ -209,7 +216,7 @@ my_new_from_file_at_size (const char *filename,
 			if (!gdk_pixbuf_loader_write (loader, buffer, length, error)) {
 			        gdk_pixbuf_loader_close (loader, NULL);
 				fclose (f);
-				g_object_unref (G_OBJECT (loader));
+				g_object_unref (loader);
 				return NULL;
 			}
 	}
@@ -218,14 +225,14 @@ my_new_from_file_at_size (const char *filename,
 
 	g_assert (*error == NULL);
 	if (!gdk_pixbuf_loader_close (loader, error)) {
-		g_object_unref (G_OBJECT (loader));
+		g_object_unref (loader);
 		return NULL;
 	}
 
 	pixbuf = gdk_pixbuf_loader_get_pixbuf (loader);
 
 	if (!pixbuf) {
-		g_object_unref (G_OBJECT (loader));
+		g_object_unref (loader);
 
 		/* did the loader set an error? */
 		if (*error != NULL)
@@ -241,7 +248,7 @@ my_new_from_file_at_size (const char *filename,
 
 	g_object_ref (pixbuf);
 
-	g_object_unref (G_OBJECT (loader));
+	g_object_unref (loader);
 
 	return pixbuf;
 }
@@ -368,6 +375,25 @@ set_filename_existing_nonexistent_cb (GtkButton      *button,
   set_filename (chooser, "/usr/nonexistent");
 }
 
+static void
+kill_dependent (GtkWindow *win, GtkObject *dep)
+{
+  gtk_object_destroy (dep);
+  g_object_unref (dep);
+}
+
+static void
+notify_multiple_cb (GtkWidget  *dialog,
+		    GParamSpec *pspec,
+		    GtkWidget  *button)
+{
+  gboolean multiple;
+
+  multiple = gtk_file_chooser_get_select_multiple (GTK_FILE_CHOOSER (dialog));
+  
+  gtk_widget_set_sensitive (button, multiple);
+}
+
 int
 main (int argc, char **argv)
 {
@@ -384,6 +410,10 @@ main (int argc, char **argv)
   
   gtk_init (&argc, &argv);
 
+  /* to test rtl layout, set RTL=1 in the environment */
+  if (g_getenv ("RTL"))
+    gtk_widget_set_default_direction (GTK_TEXT_DIR_RTL);
+  
   action = GTK_FILE_CHOOSER_ACTION_OPEN;
 
   /* lame-o arg parsing */
@@ -403,7 +433,7 @@ main (int argc, char **argv)
 
   dialog = g_object_new (GTK_TYPE_FILE_CHOOSER_DIALOG,
 			 "action", action,
-			 /* "file-system-backend", "gtk+",*/
+			 "file-system-backend", "gtk+",
 			 "select-multiple", multiple,
 			 NULL);
   switch (action)
@@ -447,6 +477,9 @@ main (int argc, char **argv)
   gtk_file_filter_add_mime_type (filter, "image/png");
   gtk_file_chooser_add_filter (GTK_FILE_CHOOSER (dialog), filter);
 
+  g_signal_connect (dialog, "notify::filter", 
+		    G_CALLBACK (filter_changed), NULL);
+
   /* Make this filter the default */
   gtk_file_chooser_set_filter (GTK_FILE_CHOOSER (dialog), filter);
   
@@ -456,11 +489,16 @@ main (int argc, char **argv)
   gtk_file_filter_add_mime_type (filter, "image/png");
   gtk_file_chooser_add_filter (GTK_FILE_CHOOSER (dialog), filter);
 
+  filter = gtk_file_filter_new ();
+  gtk_file_filter_set_name (filter, "Images");
+  gtk_file_filter_add_pixbuf_formats (filter);
+  gtk_file_chooser_add_filter (GTK_FILE_CHOOSER (dialog), filter);
+  
   /* Preview widget */
   /* THIS IS A TERRIBLE PREVIEW WIDGET, AND SHOULD NOT BE COPIED AT ALL.
    */
   preview_vbox = gtk_vbox_new (0, FALSE);
-  /*  gtk_file_chooser_set_preview_widget (GTK_FILE_CHOOSER (dialog), preview_vbox);*/
+  /*gtk_file_chooser_set_preview_widget (GTK_FILE_CHOOSER (dialog), preview_vbox);*/
   
   preview_label = gtk_label_new (NULL);
   gtk_box_pack_start (GTK_BOX (preview_vbox), preview_label, TRUE, TRUE, 0);
@@ -503,6 +541,8 @@ main (int argc, char **argv)
   gtk_container_add (GTK_CONTAINER (vbbox), button);
   g_signal_connect_swapped (button, "clicked",
 			    G_CALLBACK (gtk_file_chooser_select_all), dialog);
+  g_signal_connect (dialog, "notify::select-multiple", 
+		    G_CALLBACK (notify_multiple_cb), button);
   
   button = gtk_button_new_with_mnemonic ("_Unselect all");
   gtk_container_add (GTK_CONTAINER (vbbox), button);
@@ -514,24 +554,34 @@ main (int argc, char **argv)
   g_signal_connect (button, "clicked",
 		    G_CALLBACK (set_folder_nonexistent_cb), dialog);
 
-  button = gtk_button_new_with_label ("set_current_folder (\"/usr/nonexistent\"");
+  button = gtk_button_new_with_label ("set_current_folder (\"/usr/nonexistent\")");
   gtk_container_add (GTK_CONTAINER (vbbox), button);
   g_signal_connect (button, "clicked",
 		    G_CALLBACK (set_folder_existing_nonexistent_cb), dialog);
 
-  button = gtk_button_new_with_label ("set_filename (\"/nonexistent\"");
+  button = gtk_button_new_with_label ("set_filename (\"/nonexistent\")");
   gtk_container_add (GTK_CONTAINER (vbbox), button);
   g_signal_connect (button, "clicked",
 		    G_CALLBACK (set_filename_nonexistent_cb), dialog);
 
-  button = gtk_button_new_with_label ("set_filename (\"/usr/nonexistent\"");
+  button = gtk_button_new_with_label ("set_filename (\"/usr/nonexistent\")");
   gtk_container_add (GTK_CONTAINER (vbbox), button);
   g_signal_connect (button, "clicked",
 		    G_CALLBACK (set_filename_existing_nonexistent_cb), dialog);
 
   gtk_widget_show_all (control_window);
-  
+
+  g_object_ref (control_window);
+  g_signal_connect (dialog, "destroy",
+		    G_CALLBACK (kill_dependent), control_window);
+
+  /* We need to hold a ref until we have destroyed the widgets, just in case
+   * someone else destroys them.  We explicitly destroy windows to catch leaks.
+   */
+  g_object_ref (dialog);
   gtk_main ();
+  gtk_widget_destroy (dialog);
+  g_object_unref (dialog);
 
   return 0;
 }
